SELECT
    2 BETWEEN 1 AND 3, 2 BETWEEN 3 AND 1, 2 NOT BETWEEN 1 AND 3,
    2 BETWEEN SYMMETRIC 3 AND 1, 2 NOT BETWEEN SYMMETRIC 3 AND 1,
    1 IS DISTINCT FROM NULL, NULL IS DISTINCT FROM NULL,
    1 IS NOT DISTINCT FROM NULL, NULL IS NOT DISTINCT FROM NULL, 1.5 IS NULL,
    'null' IS NOT NULL, true IS TRUE, NULL::boolean IS TRUE, false IS NOT TRUE,
    NULL::boolean IS NOT TRUE, true IS FALSE, NULL::boolean IS FALSE,
    true IS NOT FALSE, NULL::boolean IS NOT FALSE, true IS UNKNOWN,
    NULL::boolean IS UNKNOWN, true IS NOT UNKNOWN, NULL::boolean IS NOT UNKNOWN,
    'a' LIKE '%', 'a' ILIKE '%', NULLIF('a', 'a'), 'a' SIMILAR TO '%'
