// Generated by ./scripts/update_summary_tests_list.rb

&it_parses_simple_query,
&it_parses_simple_query_with_alias,
&it_parses_query_with_nested_select_where,
&it_handles_errors,
&it_handles_basic_query,
&it_handles_join_expression,
&it_handles_recursion_without_error,
&it_parses_real_queries,
&it_parses_empty_queries,
&it_parses_floats_with_leading_dot,
&it_parses_bit_strings_hex_notation,
&it_parses_ALTER_TABLE,
&it_parses_SET,
&it_parses_SHOW,
&it_parses_COPY,
&it_parses_DROP_TABLE,
&it_parses_COMMIT,
&it_parses_CHECKPOINT,
&it_parses_VACUUM,
&it_parses_MERGE,
&it_parses_EXPLAIN,
&it_parses_SELECT_INTO,
&it_parses_LOCK,
&it_parses_CREATE_TABLE,
&it_parses_CREATE_TABLE_AS,
&it_fails_to_parse_CREATE_TABLE_WITH_OIDS,
&it_parses_CREATE_INDEX,
&it_parses_CREATE_SCHEMA,
&it_parses_CREATE_VIEW,
&it_parses_REFRESH_MATERIALIZED_VIEW,
&it_parses_CREATE_RULE,
&it_parses_CREATE_TRIGGER,
&it_parses_DROP_SCHEMA,
&it_parses_DROP_VIEW,
&it_parses_DROP_INDEX,
&it_parses_DROP_RULE,
&it_parses_DROP_TRIGGER,
&it_parses_GRANT,
&it_parses_REVOKE,
&it_parses_TRUNCATE,
&it_parses_WITH,
&it_parses_multi_line_functions,
&it_parses_table_functions,
&it_finds_called_functions,
&it_finds_functions_invoked_with_CALL,
&it_finds_dropped_functions,
&it_finds_renamed_functions,
&it_finds_nested_tables_in_SELECT,
&it_separates_CTE_names_from_table_names,
&it_finds_tables_in_SELECT_FROM_TABLESAMPLE,
&it_finds_tables_in_SELECT_FROM_XMLTABLE,
&it_ignores_JSON_TABLE,
&it_finds_nested_tables_in_FROM_clause,
&it_finds_nested_tables_in_WHERE_clause,
&it_finds_tables_in_SELECT_with_subselects_without_FROM,
&it_finds_nested_tables_in_IN_clause,
&it_finds_nested_tables_in_ORDER_BY_clause,
&it_finds_nested_tables_in_ORDER_BY_clause_with_multiple_entries,
&it_finds_nested_tables_in_GROUP_BY_clause,
&it_finds_nested_tables_in_GROUP_BY_clause_with_multiple_entries,
&it_finds_nested_tables_in_HAVING_clause,
&it_finds_nested_tables_in_HAVING_clause_with_boolean_expression,
&it_finds_nested_tables_in_a_subselect_on_a_JOIN,
&it_finds_nested_tables_in_a_subselect_in_a_JOIN_condition,
&it_correctly_categorizes_CTEs_after_UNION_SELECT,
&it_correctly_categorizes_CTEs_after_EXCEPT_SELECT,
&it_correctly_categorizes_CTEs_after_INTERSECT_SELECT,
&it_finds_tables_inside_subselectes_in_MIN_MAX_COALESCE,
&it_finds_tables_inside_CASE_statements,
&it_finds_tables_inside_casts,
&it_finds_functions_in_FROM_clause,
&it_finds_functions_in_LATERAL_clause,
&it_parses_INSERT,
&it_parses_UPDATE,
&it_parses_DELETE,
&it_parses_DROP_TYPE,
&it_finds_unqualified_names,
&it_finds_qualified_names,
&it_traverses_into_ctes,
&it_recognizes_boolean_tests,
&it_recognizes_null_tests,
&it_finds_coalesce_argument_names,
&it_finds_unqualified_names_in_union_query,
&it_finds_unqualified_names_in_union_all_query,
&it_finds_unqualified_names_in_except_query,
&it_finds_unqualified_names_in_except_all_query,
&it_finds_unqualified_names_in_intersect_query,
&it_finds_unqualified_names_in_intersect_all_query,
&it_ignores_target_list_columns,
&it_ignores_order_by_columns,
&it_handles_all_tables_in_schema,
&it_handles_schema_qualified_columns,
